"""
Script to automatically add potential trees in a given region, along roads and paths.

Road information is downloaded from OpenStreetMap.

Trees are exported in a CSV table, a PNG diagram and an HTML interactive map.
"""
import pickle
from pathlib import Path
from collections import namedtuple

import folium
import kdtree
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import numpy as np
import overpy
from pyproj import Transformer
from shapely import LineString, geometry, wkt
from shapely.ops import transform

from import_existing_trees import get_existing_trees

# TODO: Use Args
# TODO: Document
# TODO: Write issue
# TODO: Write tests?
# TODO: Export shapefile?

# From RegionChooser, or https://transfer.hft-stuttgart.de/gitlab/circulargreensimcity/circulargreensimcity/-/wikis/Fallstudien/Gromb%C3%BChl
WKT = "POLYGON((9.947021 49.803063, 9.947011 49.800917, 9.955025 49.800810, 9.955110 49.803019, 9.947021 49.803063))"
EXISTING_TREES = 'Trees_ideal_2_20240227/Trees_ideal_2_20240227.shp'
EXISTING_TREES = None
# WKT = "POLYGON((9.170419 48.782366, 9.170032 48.780825, 9.169904 48.780401, 9.170440 48.778733, 9.176877 48.780118, 9.177006 48.781193, 9.177049 48.782564, 9.176298 48.782593, 9.175440 48.782409, 9.174646 48.783399, 9.170419 48.782366))"
# Fellbach
# WKT = "POLYGON((9.271353 48.811327, 9.271911 48.809010, 9.272147 48.807187, 9.275838 48.807173, 9.275602 48.806749, 9.276138 48.806325, 9.277683 48.806424, 9.277319 48.812514, 9.275581 48.811991, 9.271353 48.811327))"
EPSG_ID = 25832
# Trees will be planted every TREE_DISTANCE along roads:
TREE_DISTANCE = 10  # [m]
# Unless there's already another tree closer than MIN_DISTANCE away:
MIN_DISTANCE = TREE_DISTANCE * 0.5  # [m]
# For display purposes only:
GRID = 100  # [m]

IGNORE_ROADS = set(['primary', 'unclassified', 'secondary',
                    'secondary_link', 'trunk', 'trunk_link', 'primary_link'])


SCRIPT_DIR = Path(__file__).resolve().parent


def load_region(wkt_polygon):
    region = wkt.loads(wkt_polygon)
    bounds = namedtuple("Bounds", "W S E N")(*region.bounds)
    return region, bounds


def get_basename(bounds):
    return f'{bounds.S}__{bounds.N}__{bounds.W}__{bounds.E}_{TREE_DISTANCE}m'.replace('.', '_')


def get_osm_roads(bounds):
    cache_dir = SCRIPT_DIR / 'cache'
    cache_dir.mkdir(exist_ok=True)

    cache_file = (cache_dir / get_basename(bounds)).with_suffix('.pickle')

    if cache_file.exists():
        print("Cache has been found. Parsing...")
        with open(cache_file, 'rb') as cache:
            ways = pickle.load(cache)
    else:
        print("Downloading data...")
        # TODO: Could add trees from OSM or Bäumekataster too.
        api = overpy.Overpass()
        result = api.query(f"""
        way({bounds.S},{bounds.W},{bounds.N},{bounds.E}) ["highway"];
        (._;>;);
        out body;
        """)
        ways = result.ways
        print("Caching data...")
        with open(cache_file, 'wb') as cache:
            pickle.dump(ways, cache)
    return ways


def set_plot(bounds, to_local_coordinates):
    x_min, y_min = to_local_coordinates.transform(bounds.W, bounds.S)
    x_max, y_max = to_local_coordinates.transform(bounds.E, bounds.N)
    ax = plt.axes()
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)
    ax.set_aspect(1)

    x_grid = plticker.MultipleLocator(base=GRID)
    y_grid = plticker.MultipleLocator(base=GRID)
    ax.xaxis.set_major_locator(x_grid)
    ax.yaxis.set_major_locator(y_grid)
    return ax


def place_trees(existing_trees_coords, ways, region, to_local, tree_distance, min_distance_2):
    local_region = transform(to_local.transform, region)

    existing_trees = kdtree.create(existing_trees_coords)

    tree_xs = []
    tree_ys = []
    for x, y in existing_trees_coords:
        tree_xs.append(x)
        tree_ys.append(y)

    for way in ways:
        width = float(way.tags.get("width", 0))
        highway = way.tags.get("highway")
        if highway in IGNORE_ROADS:
            color = 'orange'
        else:
            color = 'gray'

        local_xy_s = [to_local.transform(node.lon, node.lat)
                      for node in way.nodes]

        tree_path = LineString(local_xy_s)

        if width:
            road_as_polygon = tree_path.buffer(width / 2, cap_style='flat')
            tree_path = road_as_polygon.exterior
            displayed_width = width
        else:
            # NOTE: Could try to guess width depending on highway type.
            displayed_width = 1

        xs, ys = zip(*local_xy_s)
        plt.plot(xs, ys, linewidth=displayed_width,
                 c=color, alpha=0.8, zorder=-1)

        distances = np.arange(0, tree_path.length, tree_distance)
        potential_trees = [tree_path.interpolate(
            distance) for distance in distances]
        if tree_path.boundary:
            potential_trees += [tree_path.boundary.geoms[-1]]

        for potential_tree in potential_trees:
            x = potential_tree.x
            y = potential_tree.y
            if local_region.contains(geometry.Point(x, y)):
                _nearest_tree, distance_2 = existing_trees.search_nn((x, y))
                if distance_2 > min_distance_2:
                    existing_trees.add((x, y))
                    tree_xs.append(x)
                    tree_ys.append(y)
    return tree_xs, tree_ys


def plot_trees(bounds, tree_xs, tree_ys, tree_distance):
    plt.scatter(tree_xs, tree_ys, s=2, c='green')

    plt.grid(True)
    plt.title(f"{bounds}\nTree distance : {tree_distance} m")
    plt.gcf().set_size_inches(15, 10)
    plt.savefig(
        SCRIPT_DIR / f"{get_basename(bounds)}.png", bbox_inches='tight', dpi=300)


def export_map(bounds, tree_xs, tree_ys, epsg_id):
    to_wgs84 = Transformer.from_crs(f"EPSG:{epsg_id}", "EPSG:4326", always_xy=True)
    interactive_map = folium.Map()
    interactive_map.fit_bounds([(bounds.S, bounds.W), (bounds.N, bounds.E)])

    radius = 2  # [m]
    for x, y in zip(tree_xs, tree_ys):
        lon, lat = to_wgs84.transform(x, y)
        folium.Circle(
            location=[lat, lon],
            radius=radius,
            color="black",
            weight=1,
            fill_opacity=0.9,
            opacity=1,
            fill_color="#00ff15",
            fill=False,  # gets overridden by fill_color
            popup="{} meters".format(radius),
            tooltip="I am a tree in street STREET",
        ).add_to(interactive_map)

    folium.TileLayer(
        tiles='https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}',
        attr='Esri',
        name='Esri Satellite',
        overlay=False,
        control=True
    ).add_to(interactive_map)

    interactive_map.save(f"{get_basename(bounds)}_trees.html")


def export_csv(bounds, tree_xs, tree_ys, wkt_polygon, tree_distance, min_distance, epsg_id):
    with open(SCRIPT_DIR / f"{get_basename(bounds)}_trees.csv", "w") as csv:
        csv.write(f"# Fake trees for; {wkt_polygon}\n")
        csv.write(f"# Tree distance along roads; {tree_distance}; [m]\n")
        csv.write(f"# Minimum allowed distance between trees; {min_distance}; [m]\n")
        csv.write(f"# EPSG; {epsg_id}\n")
        csv.write("# X; Y\n")
        csv.write("# [m]; [m]\n")
        for x, y in zip(tree_xs, tree_ys):
            csv.write(f"{x};{y}\n")

    print("DONE!")


def main(wkt_polygon, epsg_id, tree_distance, min_distance, import_tree_shp):
    region, bounds = load_region(wkt_polygon)
    ways = get_osm_roads(bounds)

    if import_tree_shp:
        existing_trees = get_existing_trees(import_tree_shp)
    else:
        existing_trees = []

    to_local = Transformer.from_crs("EPSG:4326", f"EPSG:{epsg_id}", always_xy=True)

    set_plot(bounds, to_local)
    tree_xs, tree_ys = place_trees(existing_trees, ways, region,
                                   to_local, tree_distance, min_distance**2)

    plot_trees(bounds, tree_xs, tree_ys, tree_distance)
    export_map(bounds, tree_xs, tree_ys, epsg_id)
    export_csv(bounds, tree_xs, tree_ys, wkt_polygon,
               tree_distance, min_distance, epsg_id)


if __name__ == "__main__":
    main(WKT, EPSG_ID, TREE_DISTANCE, MIN_DISTANCE, EXISTING_TREES)
